
"""
修正履歴
2020/05/27
出力されるExcelファイルの列の幅を調整
write_result_xlsx() 内に処理を追加　(2020/05/27 のコメント文参照)
"""
# 変位法による立体骨組の応力解析
# 最初にクラス Stiff3D の定義があります
# これを呼び出すメインルーチンは本ファイルの末尾にあります

import math
import openpyxl
from openpyxl.utils import get_column_letter # 列幅の指定 2020/05/27

class Stiff3D:
    
    def __init__(self, inp_path = 'data.xlsx', out_path = 'result.xlsx'):
        self.inp_path = inp_path # 入力ファイルのパス
        self.out_path = out_path # 出力ファイルのパス
        self.node = []
        self.memb = []
        self.node_load = []
        self.memb_load = []
        self.node_index = [] # node_index[i] 節点番号iが定義されている node[]のindex
        self.memb_index = [] # memb_index[i] 部材番号iが定義されている memb[]のindex
        self.error = "" # エラーメッセージ
        
    def add_node(self, id = 0, x = 0, y = 0, z = 0, x_fix = 0, y_fix = 0, 
                         z_fix = 0, rx_fix = 0, ry_fix = 0, rz_fix = 0):
        # 節点情報を一つずつ追加
        # id:節点番号, x:X座標, y:Y座標, z:Z座標,
        # x_fix, y_fix, z_fix:1の時、X/Y/Z方向固定
        # rx_fix, ry_fix, rz_fix:1の時、X軸回り/Y軸回り/Z軸回り固定
        # node[i][j]
        # j = 0: id, 1: x, 2: y, 3:z, 4:x_fix, 5: y_fix, 6: z_fix, 
        #     7: rx_fix, 8: ry_fix, 9:rz_fix, 
        #     計算結果 
        #     10: X方向変位, 11: Y方向変位,  12: Z方向変位,
        #     13: X回り回転変位, 14: Y回り回転変位, 15: Z回り回転変位
        data = [0 for i in range(16)]
        data[0] = id
        data[1] = 100 * x # m -> cm
        data[2] = 100 * y
        data[3] = 100 * z
        data[4] = x_fix
        data[5] = y_fix
        data[6] = z_fix
        data[7] = rx_fix
        data[8] = ry_fix
        data[9] = rz_fix
        self.node.append(data)
        
    def set_all_node(self, list2):
        # 節点情報を list2[][] から一括読み込み
        # 戻り値 0: 失敗, 1: 成功
        if len(list2) > 1: # list2[0]はヘッダー
            try: 
                for i in range(1, len(list2)):
                    data = [0 for i in range(16)]
                    for j in range(10):
                        if list2[i][j] is not None:
                            data[j] = list2[i][j]
                    for k in range(1, 4):
                        data[k] *= 100 # m -> cm
                    self.node.append(data)
            except:
                return 0
        return 1
                
    def add_memb(self, id = 0, node1 = 0, node2 = 0, code = 0, a = 0, ix = 0, 
                 iy = 0, iz = 0, xz_as = 0, xy_as = 0, e = 0, g = 0, x_pin1 = 0, 
                 x_pin2 = 0, y_pin1 = 0, y_pin2 = 0, z_pin1 = 0, z_pin2 = 0, 
                 xz_rigid1 = 0, xz_rigid2 = 0, xy_rigid1 = 0, xy_rigid2 = 0):
        # 部材情報の追加 断面性能の単位 cm
        # id:部材番号, node1:始端の節点番号, node2:終端の節点番号, code:コードアングル,
        # a:A, ix:x軸回りのI, iy:y軸回りのI, iz:z軸回りのI, 
        # xz_as:xz面内のAs, xy_as:xy面内のAs, 
        # e:E, g:G, x_pin1, x_pin2:始端・終端のx軸回り接合(1の時ピン)
        # y_pin1, y_pin2:始端・終端のy軸回り接合(1の時ピン)
        # z_pin1, z_pin2:始端・終端のz軸回り接合(1の時ピン)
        # xz_rigid1, xz_rigid2: 始端・終端のxz面内の剛域長(m)
        # xy_rigid1, xy_rigid2: 始端・終端のxy面内の剛域長(m)
        # memb[i][j]
        # j = 0: id, 1: node1, 2: node2, 3:code,  4: A, 5: Ix, 6: Iy, 7, Iz,
        #     8: Asy, 9: Asz, 10: E, 11: G, 12: x_pin1, 13: x_pin2, 
        #     14: y_pin1, 15: y_pin2, 16: z_pin1, 17: z_pin2, 18: y_rigid1, 
        #     19: y_rigid2, 20: z_rigid1, 21: z_rigid2,
        #     計算結果 
        #     22: Ni, 23: Nj, 24: Mxi, 25: Mxj, 26: Myi, 27: Myc, 28: Myj, 
        #     29: Qyi, 30: Qyj, 31: Mzi, 32: Mzc, 33: Mzj, 34: Qzi, 35: Qzj 
        data = [0 for i in range(36)]
        data[0] = id
        data[1] = node1
        data[2] = node2
        data[3] = math.pi * code / 180 # 度 -> rad
        data[4] = a
        data[5] = ix
        data[6] = iy
        data[7] = iz
        data[8] = xz_as
        data[9] = xy_as
        data[10] = 0.1 * e # N/mm2 -> kN/cm2
        data[11] = 0.1 * g # N/mm2 -> kN/cm2
        data[12] = x_pin1
        data[13] = x_pin2
        data[14] = y_pin1
        data[15] = y_pin2
        data[16] = z_pin1
        data[17] = z_pin2
        data[18] = 100 * xz_rigid1 # m -> cm
        data[19] = 100 * xz_rigid2
        data[20] = 100 * xy_rigid1
        data[21] = 100 * xy_rigid2
        self.memb.append(data)

    def set_all_memb(self, list2):
        # 部材情報を list2[][] から一括読み込み
        # 戻り値 0: 失敗, 1: 成功
        if len(list2) > 1:  # list2[0]はヘッダー
            try: 
                for i in range(1, len(list2)):
                    data = [0 for i in range(36)]
                    for j in range(22):
                        if list2[i][j] is not None:
                            data[j] = list2[i][j]
                    data[3] = math.pi * data[3] / 180 # 度 -> rad
                    data[10] *= 0.1 # N/mm2 -> kN/cm2
                    data[11] *= 0.1 # N/mm2 -> kN/cm2
                    for k in range(18, 4):
                        data[k] *= 100 # m -> cm
                    self.memb.append(data)
            except:
                return 0
        return 1
        
    def add_node_load(self, num = 0, px = 0, py = 0, pz = 0,
                                          mx = 0, my = 0, mz = 0):
        # 節点荷重情報の追加
        # num: 節点番号, px:X方向荷重(kN), py:Y方向荷重, pz:Z方向荷重, 
        # mx:X軸回り曲げ(kN.m), my:Y軸回り曲げ, mz:Z軸回り曲げ 
        # node_load[i][j]
        # j = 0: num, 1: px, 2: py, 3: pz, 4: mx, 5: my, 6: mz 
        data = [0 for i in range(7)]
        data[0] = num
        data[1] = px
        data[2] = py
        data[3] = pz
        data[4] = 100 * mx # kN.m -> kN.cm
        data[5] = 100 * my
        data[6] = 100 * mz
        self.node_load.append(data)

    def set_all_node_load(self, list2):
        # 節点荷重情報を list2[][] から一括読み込み
        # 戻り値 0: 失敗, 1: 成功
        if len(list2) > 1: # list2[0]はヘッダー
            try: 
                for i in range(1, len(list2)):
                    data = [0 for i in range(7)]
                    for j in range(7):
                        if list2[i][j] is not None:
                            data[j] = list2[i][j]
                    data[4] *= 100 # kN.m -> kN.cm
                    data[5] *= 100
                    data[6] *= 100
                    self.node_load.append(data)
            except:
                return 0
        return 1

    def add_memb_load(self, num = 0, xz_ci = 0, xz_cj = 0, xz_m0 = 0, xz_qi = 0, 
                      xz_qj = 0, xy_ci = 0, xy_cj = 0, xy_m0 = 0, xy_qi = 0,
                      xy_qj = 0, ni = 0, nj = 0):
        # 部材荷重情報の追加
        # num:部材番号, 
        # xz_ci:i端xz面の固定端モーメント(kN.m), xz_cj:j端xz面の固定端モーメント
        # xz_m0:単純梁のxz面中央モーメント, xz_qi:i端xz面の固定端せん断力(kN), 
        # xz_qj:j端xz面の固定端せん断力, 
        # xy_ci:i端xy面の固定端モーメント(kN.m), xy_cj:j端xy面の固定端モーメント
        # xy_m0:単純梁のxy面中央モーメント, xy_qi:i端xy面の固定端せん断力(kN), 
        # xy_qj:j端xy面の固定端せん断力, 
        # ni: i端の固定端軸力(kN), nj: j端の固定端軸力(kN)
        # memb_load[i][j]
        # j = 0: num, 1: xz_ci, 2: xz_cj, 3: xz_m0, 4: xz_qi, 5: xz_qj, 
        #     6: xy_ci, 7: xy_cj, 8: xy_m0, 9: xy_qi, 10: xy_qj, 
        #     11: ni, 12: nj
        data = [0 for i in range(13)]
        data[0] = num
        data[1] = 100 * xz_ci # kN.m -> kN.cm
        data[2] = 100 * xz_cj
        data[3] = 100 * xz_m0
        data[4] = xz_qi
        data[5] = xz_qj
        data[6] = 100 * xy_ci # kN.m -> kN.cm
        data[7] = 100 * xy_cj
        data[8] = 100 * xy_m0
        data[9] = xy_qi
        data[10] = xy_qj
        data[11] = ni
        data[12] = nj
        self.memb_load.append(data)
        
    def set_all_memb_load(self, list2):
        # 部材荷重情報を list2[][] から一括読み込み
        # 戻り値 0: 失敗, 1: 成功
        if len(list2) > 1: # list2[0]はヘッダー
            try: 
                for i in range(1, len(list2)):
                    data = [0 for i in range(13)]
                    for j in range(13):
                        if list2[i][j] is not None:
                            data[j] = list2[i][j]
                    for k in range(3):
                        data[1 + k] *= 100 # kN.m -> kN.cm
                        data[6 + k] *= 100
                    self.memb_load.append(data)
            except:
                return 0
        return 1

    def calc_main(self):

        if not self.pre_check():
            return 0
        # 自由度(未知数)に番号をふる
        # free[i][j]: 節点node[i]の自由度番号 
        #         j = 0:X方向, 1:Y方向, 2:Z方向, 3:X軸回り, 4:Y軸回り, 5:Z軸回り 
        free = []
        num_free = -1
        for i in range(len(self.node)):
            free1 = [-1 for i in range(6)]
            for j in range(6):
                if self.node[i][j + 4] == 0:
                    num_free += 1
                    free1[j] = num_free
            free.append(free1)
            
        max_free = num_free + 1 # 自由度(未知数)の数 = 全体剛性マトリクスのサイズ
        if max_free <= 0:
            self.error = "自由度が存在しない"
            return 0
    
        # バンドマトリクスの作成
        # counts[i]: バンドマトリクスのi列にある要素の数
        counts = [0 for i in range(max_free)]
        for m in range(len(self.memb)):
            for i in range(2): # i端/j端
                na = self.node_index[self.memb[m][i + 1]]
                for j in range(2):
                    nb = self.node_index[self.memb[m][j + 1]]
                    for k in range(6): # X/Y/Z/XR/YR/ZR
                        for n in range(6): # X/Y/Z/XR/YR/ZR
                            ndi = free[na][k]
                            ndj = free[nb][n]
                            if (ndi != -1) and (ndj != -1):
                                if (ndj - ndi + 1) > counts[ndj]:
                                    counts[ndj] = ndj - ndi + 1

        #  diag_num[i]: バンドマトリクスi列の対角要素の番号
        # 全体剛性マトリクスの要素[i,j]のバンドマトリクスの要素番号は diag_num[j]+j-i
        diag_num = [0 for i in range(max_free + 1)]
        num = 0
        for i in range(1, max_free + 1):
            num += counts[i - 1]
            diag_num[i] = num

        # 全体剛性マトリクスのバンドマトリクス band[]
        band = self.global_matrix(free, diag_num, max_free)
        # band[] を上三角行列に変換
        if not self.pre_gauss(band, diag_num, num_free):
            return 0
        # cmq_index[i] : memb[i] に荷重項の指定がある場合の memb_load[] のindex
        cmq_index = [-1 for i in range(len(self.memb))]
        # 外力ベクトル load[] の作成
        load = self.load_vector(free, num_free, cmq_index)
        # 変位ベクトル disp[] の計算
        disp = self.solve_gauss(band, diag_num, num_free, load)
        # 節点変位量を node[] に格納
        self.set_node_disp(free, disp)
        # 部材応力を memb[] に格納
        self.set_memb_stress(free, disp, cmq_index)
        return 1

    def global_matrix(self, free, diag_num, max_free):
        # 戻り値 全体剛性マトリクスのバンドマトリクス
        band = [0 for i in range(diag_num[max_free])] # 全体剛性マトリクスのバンドマトリクス
        free1 = 6 # 自由度数
        free2 = 12 # 全自由度数
        wk =  [[0 for i in range(free2)] for j in range(free2)]
        gk =  [[0 for i in range(free2)] for j in range(free2)]
        for m in range(len(self.memb)):
            stiff = self.memb_matrix(m) # 部材剛性マトリクス [12][12]
            trans = self.trans_matrix(m) # 座標変換マトリクス [12][12]
            # wk[][] = trans[][]T * stiff[][]
            for i in range(free2):
                for j in range(free2):
                    s = 0
                    for k in range(free2):
                        s += trans[k][i] * stiff[k][j]
                    wk[i][j] = s
            # gk[][] = stiff[][] * trans[][] : 全体座標に変換したマトリクス
            for i in range(free2):
                for j in range(free2):
                    s = 0
                    for k in range(free2):
                        s += wk[i][k] * trans[k][j]
                    gk[i][j] = s
            # 部材剛性マトリクス stiff[][] を band[] に組み込む
            for i in range(2): # i端/j端
                na = self.node_index[self.memb[m][i + 1]]
                for j in range(2):
                    nb = self.node_index[self.memb[m][j + 1]]
                    for k in range(free1): # X/Y/Z/XR/YR/ZR
                        for n in range(free1): # X/Y/Z/XR/YR/ZR
                            ndi = free[na][k]
                            ndj = free[nb][n]
                            if (ndj >= ndi) and (ndi != -1) and (ndj != -1):
                                band[diag_num[ndj] + ndj - ndi] += \
                                    gk[free1 * i + k][free1 * j + n]

        for i in range(len(diag_num) - 1):
            if abs(band[diag_num[i]]) < 0.001:
                # 対角要素が0の場合は1を代入して計算を続行
                band[diag_num[i]] = 1

        return band        
            
    def memb_matrix(self, m):
        # 戻り値 memb[m]の部材マトリクス[12][12]
        # code = 0: 部材座標系, 1: 全体座標系
        free2 = 12 # 全自由度数
        stiff =  [[0 for i in range(free2)] for j in range(free2)]
        al = self.memb_length(m, 1)
        # y軸回り(xz面)の剛性係数
        lay = self.memb[m][18] # 始端剛域長
        lby = self.memb[m][19] # 終端剛域長
        fk1 = self.memb_stiff(l = al, e = self.memb[m][10], g = self.memb[m][11], 
                    sa = self.memb[m][8], gi = self.memb[m][6], rgd1 = lay, 
                    rgd2 = lby, pin1 = self.memb[m][14], pin2 = self.memb[m][15])
        ky11 =  fk1[0]
        ky21 = -fk1[1]
        ky22 =  fk1[2]
        ky41 = -fk1[6]
        ky42 =  fk1[7]
        ky44 =  fk1[9]
        # z軸回り(xy面)の剛性係数
        laz = self.memb[m][20] # 始端剛域長
        lbz = self.memb[m][21] # 終端剛域長
        fk2 = self.memb_stiff(l = al, e = self.memb[m][10], g = self.memb[m][11], 
                    sa = self.memb[m][9], gi = self.memb[m][7], rgd1 = laz, 
                    rgd2 = lbz, pin1 = self.memb[m][16], pin2 = self.memb[m][17])
        kz11 = fk2[0]
        kz21 = fk2[1]
        kz22 = fk2[2]
        kz41 = fk2[6]
        kz42 = fk2[7]
        kz44 = fk2[9]
        stiff[0][0] = self.memb[m][10] * self.memb[m][4] / al
        stiff[1][1] = kz11
        stiff[2][2] = ky11
        if self.memb[m][12] == 0:
            stiff[3][3] = self.memb[m][11] * self.memb[m][5] / al
        stiff[4][2] = ky21 - ky11 * lay
        stiff[4][4] = ky22 - 2 * ky21 * lay + ky11 * lay * lay
        stiff[5][1] = kz21 + kz11 * laz
        stiff[5][5] = kz22 + 2 * kz21 * laz + kz11 * laz * laz
        stiff[6][0] = -stiff[0][0]
        stiff[6][6] =  stiff[0][0]
        stiff[7][1] = -stiff[1][1]
        stiff[7][5] = -stiff[5][1]
        stiff[7][7] =  stiff[1][1]
        stiff[8][2] = -stiff[2][2]
        stiff[8][4] = -stiff[4][2]
        stiff[8][8] =  stiff[2][2]
        if (self.memb[m][12] == 0) and (self.memb[m][13] == 0):
            stiff[9][3] = stiff[3][3]
        if self.memb[m][13] == 0:
            stiff[9][9] = self.memb[m][11] * self.memb[m][5] / al
        stiff[10][2] = ky41 - ky11 * lby
        stiff[10][4] = ky42 - ky41 * lay - ky21 * lby
        stiff[10][4] = stiff[10][4]+ ky11 * lay * lby
        stiff[10][8] = -stiff[10][2]
        stiff[10][10] = ky44 - 2 * ky41 * lby + ky11 * lby * lby
        stiff[11][1] = kz41 + kz11 * lbz
        stiff[11][5] = kz42 + kz41 * laz + kz21 * lbz
        stiff[11][5] = stiff[11][5] + kz11 * laz * lbz
        stiff[11][7] = -stiff[11][1]
        stiff[11][11] = kz44 + 2 * kz41 * lbz + kz11 * lbz * lbz
        for i in range(free2):
            for j in range(i + 1, free2):
                stiff[i][j] = stiff[j][i]
                
        return stiff    

    def memb_stiff(self, l, e, g, sa, gi, rgd1, rgd2, pin1, pin2):
        # 戻り値 xz面またはxy面における部材の剛性係数マトリクス fk[10]
        # l:部材長, e:ヤング係数, g:せん断断線係数, sa:せん断断面積, 
        # rgd1・rgd2:始端・終端の剛域長, pin1・pin2:始端・終端の接合状態(1の時ピン)
        fk = [0 for i in range(10)]
        ld = l - rgd1 - rgd2 # 可撓長さ
        ei = e * gi # EI
        rza = 1 # i端の接合状態による係数
        rzb = 1 # j端の接合状態による係数
        if pin1 == 1:
            rza = 0
        if pin2 == 1:
            rzb = 0
        rys = 1 # せん断剛性の係数
        if sa > 0.1:
            ks = g * sa / ld
            rys = ks / (ks + 12 * ei / (ld ** 3))
        rz = rzb + 2 * rys + rzb * rys + rza * (1 - 4 * rzb * (rys - 1) + rys)
        kz11 = (12 * ei / (ld ** 3)) * rys * (rza + rzb + 4 * rza * rzb) / rz
        kz21 = (12 * ei / (ld * ld)) * rza * rys * (1 + 2 * rzb) / rz
        kz22 = (6 * ei / ld) * rza * (rzb + 2 * rys + rzb * rys) / rz
        kz41 = (12 * ei / (ld * ld)) * rzb * rys * (1 + 2 *rza) / rz 
        kz42 = (6 * ei / ld) * rza * rzb * ( -1 + 3 * rys) / rz
        kz44 = (6 * ei / ld) * rzb * (rza + 2 * rys + rza * rys) / rz
        fk[0] =  kz11
        fk[1] =  kz21
        fk[2] =  kz22
        fk[3] = -kz11
        fk[4] = -kz21
        fk[5] =  kz11
        fk[6] =  kz41
        fk[7] =  kz42
        fk[8] = -kz41
        fk[9] =  kz44
        
        return fk

    def trans_matrix(self, m, result = 0):
        # 戻り値 部材の座標変換マトリクス [12][12]
        #       result = 1の時は方向余弦のマトリクス[3][3]
        # m:部材のindex
        tr = [[0 for i in range(3)] for j in range(3)]
        # [ lx, mx, nx ] : 部材x軸の方向余弦
        # [ ly, my, ny ] : 部材y軸の方向余弦
        # [ lz, mz, nz ] : 部材z軸の方向余弦
        (al, x1, y1, z1) = self.memb_length(m)
        tr[0][0] = x1 / al # lx
        tr[0][1] = y1 / al # mx
        tr[0][2] = z1 / al # nx
        cs = math.cos(self.memb[m][3]) # コードアングル
        sn = math.sin(self.memb[m][3])
        if (abs(x1) < 0.01) and (abs(y1) < 0.01):
            # x軸が基準Z軸に平行
            tr[1][0] =  tr[0][2] * cs # ly
            tr[1][1] =  sn # my
            tr[2][0] = -tr[0][2] * sn # lz
            tr[2][1] =  cs # mz
        else:
            rlm = math.sqrt(tr[0][0] ** 2 + tr[0][1] ** 2)
            lzd = -tr[0][1] / rlm
            mzd =  tr[0][0] / rlm
            nzd =  0
            lyd = -tr[0][0] * tr[0][2] / rlm
            myd = -tr[0][1] * tr[0][2] / rlm
            nyd = rlm
            tr[1][0] =  lzd * cs + lyd * sn # ly
            tr[1][1] =  mzd * cs + myd * sn # my
            tr[1][2] =  nzd * cs + nyd * sn # ny
            tr[2][0] = -lzd * sn + lyd * cs # lz
            tr[2][1] = -mzd * sn + myd * cs # mz
            tr[2][2] = -nzd * sn + nyd * cs # nz

        if result == 1:
            return tr
        else:
            trans = [[0 for i in range(12)] for j in range(12)]
            for i in range(3):
                for j in range(3):
                    for k in range(4):
                        trans[3 * k + i][3 * k + j] = tr[i][j]
            return trans
    
    def pre_gauss(self, a, diag_num, num_max):
        # バンドマトリクス a[] を上三角マトリクスに変換
        # diag_num[i]: a[] のi列の対角要素の番号
        # num_max: a[] の列数 - 1
        # 戻り値 0: 異常終了, 1: 正常終了
        mm = [0 for i in range(num_max + 1)]
        for i in range(num_max + 1):
            if a[diag_num[i]] < 0.01:
                self.error = "剛性マトリクスの対角要素が0になる"
                return 0
        # 上三角マトリクスの作成
        for j in range(num_max + 1):
            mm[j] = j - diag_num[j + 1] + diag_num[j] + 1
            nst = mm[j] + 1
            if j > nst:
                for i in range(nst, j):
                    mk1 = mm[i]
                    if mm[j] > mm[i]:
                        mk1 = mm[j]
                    mk2 = i - 1
                    if num_max + 1 < i:
                        mk2 = num_max
                    nij = diag_num[j] + j - i
                    for k in range(mk1, mk2 + 1):
                        a[nij] -= a[diag_num[i] + i - k] * a[diag_num[j] + j - k]
            njj = diag_num[j]
            mk2 = j - 1
            if num_max + 1 < j:
                mk2 = num_max
            if mk2 >= mm[j]:
                for i in range(mm[j], mk2 + 1):
                    nij = diag_num[j] + j - i
                    a[njj] -= a[nij] * a[nij] / a[diag_num[i]]
                    a[nij] /= a[diag_num[i]]
        return 1
    
    def solve_gauss(self, a, diag_num, num_max, r):
        # ガウス消去法による解法
        # a[]: 上三角マトリクス, diag_num[i]: a[] のi列の対角要素の番号
        # num_max: a[] の列数 - 1, r[]: 係数ベクトル
        # 戻り値 result[]: 解のベクトル
        result = []
        for i in range(len(r)):
            result.append(r[i])        
        mm = [0 for i in range(num_max + 1)]
        mm[0] = 1
        for i in range(1, num_max + 1):
            mm[i] = i - diag_num[i + 1] + diag_num[i] + 1
        for i in range(1, num_max + 1):
            for k in range(mm[i], i):
                n = diag_num[i] + i - k
                result[i] -= a[n] * result[k]
        for i in range(num_max + 1):
            result[i] /= a[diag_num[i]]                 
        for i in reversed(range(0, num_max + 1)):
            for k in range(mm[i], i):
                n = diag_num[i] + i - k
                result[k] -= a[n] * result[i]
        return result                
 
    def load_vector(self, free, num_free, cmq_index):
        # 外力ベクトルの作成
        # free[i][j]: 節点node[i]の自由度番号, num_free: 自由度の数
        # cmq_index[i]: memb[i] に対応した memb_load[] の index
        # 戻り値 外力ベクトル vect[]
        vect = [0 for i in range(num_free + 1)]
        # 節点荷重による外力
        for i in range(len(self.node_load)):
            free_num = -1
            for j in range(6): # X/Y/Z/XR/YR/ZR
                if abs(self.node_load[i][j + 1]) > 0.01:
                    num = self.node_index[self.node_load[i][0]]
                    free_num = free[num][j]
                    if free_num != -1:
                        vect[free_num] += self.node_load[i][j + 1]
        # 部材荷重による外力
        free1 = 6 # 自由度数
        free2 = 12
        for i in range(len(self.memb_load)):
            if self.memb_load[i][0] == 0:
                continue # 重複指定がある
            m = self.memb_index[self.memb_load[i][0]]
            cmq_index[m] = i
            # cmq[i][j]
            # i = 0:xz面, 1:xy面, 2:x軸回り
            # j = 0:Ci, 1:Cj, 2:M0, 3:Qi, 4:Qj
            # wk[]:cmq[][]の成分ベクトルを自由度順に並べ替え
            # (xi, yi, zi, rxi, ryi, rzi, xj, yj, zj, rxj, ryj, rzj)
            wk = [0 for j in range(free2)]
            # cmq[i] i = 0:Ci, 1:Cj, 2:M0, 3:Qi, 4:Qj
            # xz面
            cmq = self.cmq_vector(i, m, 0)
            wk[2] = cmq[3] # zi
            wk[4] = cmq[0] # θyi
            wk[8] = cmq[4] # zj
            wk[10] = cmq[1] # θyj
            # xy面
            cmq = self.cmq_vector(i, m, 1)
            wk[1] = cmq[3] # yi
            wk[5] = cmq[0] # θzi
            wk[7] = cmq[4] # yj
            wk[11] = cmq[1] # θzj
            # x軸方向            
            wk[0] = self.memb_load[i][11] # xi
            wk[6] = self.memb_load[i][12] # xj
            # 座標変換マトリクス
            trans = self.trans_matrix(m)
            # fem[]:全体座標系に変換したcmq[]の成分
            fem = [0 for j in range(free2)]
            for j in range(free2):
                s = 0
                for k in range(free2):
                    s += wk[k] * trans[k][j]
                fem[j] = s
            ni = self.node_index[self.memb[m][1]] # i端節点の node[] の index
            nj = self.node_index[self.memb[m][2]] # j端節点の node[] の index
            # num[]: 両端の自由度の番号
            num = [0 for i in range(free2)]
            for j in range(free1):
                num[j] = free[ni][j]
                num[free1 + j] = free[nj][j]
            # 外力ベクトルに加算
            for j in range(free2):
                if num[j] > -1:
                    vect[num[j]] -= fem[j];
            
        return vect
    
    def cmq_vector(self, n, m, xyz):
        # memb[m] memb_load[n] の荷重項ベクトル
        # xyz = 0:xz面, 1:xy面
        # 戻り値 vect[i] i = 0:Ci, 1:Cj, 2:M0, 3:Qi, 4:Qj
        vect = [0 for i in range(5)]
        if xyz == 0:
            for i in range(len(vect)):
                vect[i] = self.memb_load[n][1 + i]
            pin1 = self.memb[m][14]
            pin2 = self.memb[m][15]
        else:
            for i in range(len(vect)):
                vect[i] = self.memb_load[n][6 + i]
            pin1 = self.memb[m][16]
            pin2 = self.memb[m][17]

        # 接合部がピンの場合
        # pin1:i端接合部(1:ピン), pin2:j端接合部(1:ピン)
        if (pin1 > 0) or (pin2 > 0):
            al = self.memb_length(m, 1)
            (Ci, Cj, Qi, Qj) = (0, 1, 3, 4)
            if (pin1 == 1) and (pin2 == 1):
                # 両端ピン
                vect[Qi] = vect[Qi] - (vect[Ci] + vect[Cj]) / al
                vect[Qj] = vect[Qj] + (vect[Ci] + vect[Cj]) / al
                vect[Ci] = 0
                vect[Cj] = 0
            elif pin1 == 1:
                # i端ピン
                vect[Cj] = vect[Cj] - 0.5 * vect[Ci]
                vect[Qi] = vect[Qi] + 1.5 * vect[Ci] / al
                vect[Qj] = vect[Qj] - 1.5 * vect[Ci] / al
                vect[Ci] = 0
            else:
                # j端ピン
                vect[Ci] = vect[Ci] - 0.5 * vect[Cj]
                vect[Qi] = vect[Qi] + 1.5 * vect[Cj] / al
                vect[Qj] = vect[Qj] - 1.5 * vect[Cj] / al
                vect[Cj] = 0

        return vect
    
    def set_node_disp(self, free, disp):
        # 節点変位量を node[] に格納
        # free[][]: 自由度番号の配列, disp[]: 全自由度の変位ベクトル
        for i in range(len(self.node)):
            for j in range(6): # X, Y, Z, Xr, Yr, Zr
                if free[i][j] > -1:
                    if j < 3: # X, Y, Z
                        self.node[i][10 + j] = 10 * disp[free[i][j]] # cm -> mm
                    else:
                        self.node[i][10 + j] = disp[free[i][j]]

    def set_memb_stress(self, free, disp, cmq_index):
        # 部材応力を memb[] に格納
        # free[][]: 自由度番号の配列, disp[]: 全自由度の変位ベクトル
        # cmq_index[i]: memb[i] に対応した memb_load[] の index
        free1 = 6 # 自由度
        free2 = 12 # 全自由度
        # memb[][] の index 
        for m in range(len(self.memb)):
            # num[]: 部材両端の節点の自由度番号 
            #        (xi, yi, zi, rxi, ryi, rzi, xj, yj, zj, rxj, ryj, rzj)
            num = [0 for i in range(free2)]
            for i in range(free1): # X/Y/Z/rX/rY/rZ
                num[i] = free[self.node_index[self.memb[m][1]]][i] # i端側
                num[free1 + i] = free[self.node_index[self.memb[m][2]]][i] # j端側
            # global_disp[]: 部材両端の全体座標系の変位 (xi, yi, ri, xj, yj, rj)
            global_disp = [0 for i in range(free2)]
            for i in range(free2):
                if num[i] > -1:
                    global_disp[i] = disp[num[i]]
            # local_disp[]: 部材両端の部材座標系の変位
            #          (xi, yi, zi, rxi, ryi, rzi, xj, yj, zj, rxj, ryj, rzj)
            local_disp = [0 for i in range(free2)]
            # 座標変換マトリクス trans[][] (global_disp -> local_disp)
            trans = self.trans_matrix(m)
            for i in range(free2):
                s = 0
                for j in range(free2):
                    s += global_disp[j] * trans[i][j]
                local_disp[i] = s
            # local_stiff[12][12]: 部材座標系の剛性マトリクス
            local_stiff = self.memb_matrix(m)
            # stress[]: 部材応力 = local_disp * local_stiff
            stress = [0 for i in range(free2)]
            for i in range(free2):
                for j in range(free2):
                    stress[i] += local_stiff[i][j] * local_disp[j]
            m0y = 0 # 単純梁の中央曲げ(y軸回り)
            m0z = 0 # 単純梁の中央曲げ(z軸回り)
            if (cmq_index[m] > -1):
                # 荷重項の値を部材応力に加算
                # xz面 cmq[] = (ryi, ryj, m0, zi, zj)
                cmq = self.cmq_vector(cmq_index[m], m, 0)
                stress[4] += cmq[0]
                stress[10] += cmq[1]
                stress[2] += cmq[3]
                stress[8] += cmq[4]
                m0y = cmq[2]
                # xy面 cmq[] = (rzi, rzj, m0, yi, yj)
                cmq = self.cmq_vector(cmq_index[m], m, 1)
                stress[5] += cmq[0]
                stress[11] += cmq[1]
                stress[1] += cmq[3]
                stress[7] += cmq[4]
                m0z = cmq[2]
                # x軸方向
                stress[0] += self.memb_load[cmq_index[m]][11] # xi
                stress[6] += self.memb_load[cmq_index[m]][12] # xj
            # stress[]
            # 0:xi, 1:yi, 2:zi, 3:rxi, 4:ryi, 5:rzi, 6:xj, 7:yj, 8:zj, 9:rxj, 10:ryj, 11:rzj
            # stress[] -> memb[]
            # 22: Ni, 23: Nj, 24: Mxi, 25: Mxj, 26: Myi, 27: Myc, 28: Myj, 
            # 29: Qyi, 30: Qyj, 31: Mzi, 32: Mzc, 33: Mzj, 34: Qzi, 35: Qzj 
            self.memb[m][22] = stress[0]
            self.memb[m][23] = stress[6]
            self.memb[m][24] = stress[3] 
            self.memb[m][25] = stress[9]
            self.memb[m][26] = stress[4]
            self.memb[m][27] = m0y + 0.5 * ((stress[4] - stress[10]))
            self.memb[m][28] = stress[10]
            self.memb[m][29] = stress[2]
            self.memb[m][30] = stress[8]
            self.memb[m][31] = stress[5]
            self.memb[m][32] = m0z + 0.5 * ((stress[5] - stress[11]))
            self.memb[m][33] = stress[11]
            self.memb[m][34] = stress[1]
            self.memb[m][35] = stress[7]
            # kN.cm -> kN.m 
            self.memb[m][24] *= 0.01 
            self.memb[m][25] *= 0.01
            for i in range(3):
                self.memb[m][26 + i] *= 0.01
                self.memb[m][31 + i] *= 0.01

    def pre_check(self):
        # データの検証
        # 戻り値 0: エラーあり, 1: エラーなし
        if len(self.node) == 0:
            self.error = "節点情報がない"
            return 0
        elif len(self.memb) == 0:
            self.error = "部材情報がない"
            return 0
        elif (len(self.node_load) == 0) and (len(self.memb_load) == 0): 
            self.error = "荷重情報がない"
            return 0

        num_node = 0 # 節点番号の最大値
        for i in range(len(self.node)):
            if self.node[i][0] < 1:
                self.error = "節点番号" + str(self.node[i][0]) + "に誤りがある"
                return 0
            elif self.node[i][0] > num_node:
                num_node = self.node[i][0]
        # node_index[i] 節点番号iが定義されている node[]のindex
        self.node_index = [-1 for i in range(num_node + 1)]
        for i in range(len(self.node)):
            self.node_index[self.node[i][0]] = i

        num_memb = 0 # 部材番号の最大値
        for i in range(len(self.memb)):
            err_memb = "部材番号" + str(self.memb[i][0])
            if self.memb[i][0] < 1:
                self.error = err_memb + "に誤りがある"
                return 0
            elif (self.memb[i][1] < 1) or (self.memb[i][1] > num_node) or \
                    (self.memb[i][2] < 1) or (self.memb[i][2] > num_node):
                self.error = err_memb + "の節点情報に誤りがある"
                return 0
            elif (self.memb[i][4] < 0.1) or (self.memb[i][6] < 0.1) or \
                    (self.memb[i][7] < 0.1):
                self.error = err_memb + "の断面性能に誤りがある"
                return 0
            elif self.memb[i][10] < 0.1:
                self.error = err_memb + "の材料定数に誤りがある"
                return 0
            else:
                al = self.memb_length(i, 1)
                if al < 0.01:
                    self.error = err_memb + "の長さが確保できない"
                    return 0
                elif (al < (self.memb[i][16] + self.memb[i][17])) or \
                        (al < (self.memb[i][18] + self.memb[i][19])):
                    self.error = err_memb + "剛域長の値に誤りがある"
                    return 0
                else:
                    if self.memb[i][0] > num_memb:
                        num_memb = self.memb[i][0]
                
        # memb_index[i] 部材番号iが定義されている memb[]のindex
        self.memb_index = [-1 for i in range(num_memb + 1)]
        for i in range(len(self.memb)):
            self.memb_index[self.memb[i][0]] = i
        # 節点荷重情報
        for i in range(len(self.node_load)):
            if (self.node_load[i][0] < 1) or (self.node_load[i][0] > num_node):
                self.error = "節点荷重が作用する節点番号" + \
                            str(self.node_load[i][0]) + "に誤りがある"
                return 0
        # 部材荷重情報
        for i in range(len(self.memb_load)):
            if (self.memb_load[i][0] < 1) or (self.memb_load[i][0] > num_memb):
                self.error = "部材荷重が作用する部材番号" + \
                            str(self.memb_load[i][0]) + "に誤りがある"
                return 0
        # 部材荷重が重複して指定されている場合は値を加算
        for i in range(len(self.memb_load)):
            if self.memb_load[i][0] != 0: 
                for j in range(i + 1, len(self.memb_load)):
                    if self.memb_load[i][0] == self.memb_load[j][0]:
                        self.memb_load[j][0] = 0
                        for k in range(13):
                            self.memb_load[i][k] += self.memb_load[j][k]
        return 1
               
    def memb_length(self, m, result = 0):
        # memb[m]の部材長
        # result = 1: 部材長のみを返す
        # 戻り値 al:部材長, x1, y1, z1:両端のXYZ座標の差
        #       nx:部材x軸の基準Z軸に対する余弦
        x1 = self.node[self.node_index[self.memb[m][1]]][1]
        y1 = self.node[self.node_index[self.memb[m][1]]][2]
        z1 = self.node[self.node_index[self.memb[m][1]]][3]
        x2 = self.node[self.node_index[self.memb[m][2]]][1]
        y2 = self.node[self.node_index[self.memb[m][2]]][2]
        z2 = self.node[self.node_index[self.memb[m][2]]][3]
        x1 = x2 - x1
        y1 = y2 - y1
        z1 = z2 - z1
        al = x1 ** 2 + y1 ** 2 + z1 ** 2
        if al > 0.01:
            al = math.sqrt(al)
        if result:
            return al
        else:
            return (al, x1, y1, z1)
        
    def read_data_xlsx(self):
        # data.xlsx からデータを読み込む
        # 戻り値 0: 失敗, 1: 成功
        try:
            wb = openpyxl.load_workbook(self.inp_path)
            self.set_all_node(list(wb['節点'].values))
            self.set_all_memb(list(wb['部材'].values))
            self.set_all_node_load(list(wb['節点荷重'].values))
            self.set_all_memb_load(list(wb['部材荷重'].values))
            return 1        
        except Exception as err:
            print(err)
            return 0
    
    def write_result_xlsx(self):
        # result.xlsx に計算結果を書き込む
        # 戻り値 0: 失敗, 1: 成功
        def result_value(f, n):
            # 浮動小数 f を小数以下 n 桁の値に変換
            v = round((10 ** n) * abs(f)) / (10 ** n)
            if f > 0:
                return v
            else:
                return -v
        try:
            wb = openpyxl.Workbook()
            # 節点変位
            ws1 = wb.active
            ws1.title = '節点変位'
            header1 = ['節点番号', 'X方向変位(mm)', 'Y方向変位(mm)', 'Z方向変位(mm)',
                       'X軸回り変位(rad)', 'Y軸回り変位(rad)', 'Z軸回り変位(rad)']
            # node[i][j] 計算結果 
            # j = 10: X方向変位, 11: Y方向変位,  12: Z方向変位,
            #     13: X回り回転変位, 14: Y回り回転変位, 15: Z回り回転変位
            for i in range(len(header1)):
                ws1.cell(row = 1, column = i + 1).value = header1[i]
            for i in range(len(self.node)):
                result = []
                result.append(self.node[i][0])
                for j in range(3):
                    result.append(result_value(self.node[i][10 + j], 2))
                for j in range(3):
                    result.append(result_value(self.node[i][13 + j], 4))
                for j in range(len(result)):
                    ws1.cell(row = i + 2, column = j + 1).value = result[j]
            # 列幅の指定 2020/05/27
            ws1.column_dimensions[get_column_letter(1)].width = 10
            for x in range(1, len(header1)):
                ws1.column_dimensions[get_column_letter(x + 1)].width = 15
            # 部材応力
            ws2 =  wb.create_sheet('部材応力')
            header2 = ['部材番号', 'Ni (kN)', 'Nj (kN)', 'xMi (kN.m)', 'xMj (kN.m)', 
                       'xzMi (kN.m)','xzMc (kN.m)','xzMj (kN.m)', 'xzQi (kN)',
                       'xzQj (kN)', 'xyMi (kN.m)', 'xyMc (kN.m)', 'xyMj (kN.m)',
                       'xyQi (kN)', 'xyQj (kN)']
            # memb[i][j] 計算結果
            # j = 22: Ni, 23: Nj, 24: Mxi, 25: Mxj, 26: Myi, 27: Myc, 28: Myj, 
            #     29: Qyi, 30: Qyj, 31: Mzi, 32: Mzc, 33: Mzj, 34: Qzi, 35: Qzj 
            for i in range(len(header2)):
                ws2.cell(row = 1, column = i + 1).value = header2[i]
            for i in range(len(self.memb)):
                result = []
                result.append(self.memb[i][0])
                for j in range(22, 36):
                    result.append(result_value(self.memb[i][j], 2))
                for j in range(len(result)):
                    ws2.cell(row = i + 2, column = j + 1).value = result[j]
            # 列幅の指定 2020/05/27
            for x in range(len(header2)):
                ws2.column_dimensions[get_column_letter(x + 1)].width = 10
                
            wb.save(self.out_path)
            return 1        
        except Exception as err:
            print(err)
            return 0

# メインルーチン
# 入出力ファイルを初期設定以外にする場合は引数 inp_path out_path を指定
obj = Stiff3D()
# データの読み込み
if obj.read_data_xlsx():
    # 計算実行
    if obj.calc_main():
        # 結果の書き出し
        if obj.write_result_xlsx():
            print('正常終了')
        else:
            print("出力ファイルの書き込みエラー")
    else:
        print('実行時エラー : ' + obj.error)
else:
    print("入力ファイルの読み込みエラー")

